'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _staticLand = require('static-land');

var _compose = require('./compose');

var _compose2 = _interopRequireDefault(_compose);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

// We're not sure yet if this function should be public
function fromArray(xs) {
  return function (sink) {
    xs.forEach(function (x) {
      sink(x);
    });
    return function () {};
  };
}

var Stream = (0, _staticLand.fromIncomplete)({

  /* Creates an empty stream
   */

  empty: function empty() {
    return function () {
      return function () {};
    };
  },


  /* Merges several streams of same type to a single stream of that type.
   * The result stream will contain values from all streams
   */
  concat: function concat(streams) {
    return Stream.chain(function (x) {
      return x;
    })(fromArray(streams));
  },


  /* Creates a stream containing given value
   */
  of: function of(x) {
    return function (sink) {
      sink(x);
      return function () {};
    };
  },


  /* Lifts function `A => B` to a function that operates
   * on streams `Stream<A> => Stream<B>`
   */
  map: function map(fn, stream) {
    return function (sink) {
      return stream(function (payload) {
        return sink(fn(payload));
      });
    };
  },


  /* Given a stream of functions `Stream<A => B>`, returns a function
   * that operates on streams `Stream<A> => Stream<B>`.
   * The result stream `Stream<B>` will contain values created by applying
   * the latest function from `Stream<A => B>` to the latest value from `Stream<A>`
   * every time one of them updates.
   */
  ap: function ap(streamf, streamv) {
    return function (sink) {
      var latestF = { type: 'nothing' };
      var latestV = { type: 'nothing' };
      var push = function push() {
        if (latestF.type === 'just' && latestV.type === 'just') {
          var fn = latestF.value;
          sink(fn(latestV.value));
        }
      };
      var disposef = streamf(function (f) {
        latestF = { type: 'just', value: f };
        push();
      });
      var disposev = streamv(function (v) {
        latestV = { type: 'just', value: v };
        push();
      });
      return function () {
        disposef();
        disposev();
      };
    };
  },


  /* Given a function `A => Stream<B>` returns a function
   * that operates on streams `Stream<A> => Stream<B>`.
   * The result function will spawn a `Stream<B>`
   * for each value from `Stream<A>` using the provided function.
   * The final `Stream<B>` will contain values from all spawned streams.
   */
  chain: function chain(fn, stream) {
    return function (sink) {
      var spawnedDisposers = [];
      var mainDisposer = stream(function (payload) {
        spawnedDisposers.push(fn(payload)(sink));
      });
      return function () {
        spawnedDisposers.forEach(function (fn) {
          return fn();
        });
        mainDisposer();
      };
    };
  },


  /* Given an array of streams returns a stream of arrays.
   */
  combineArray: function combineArray(arr) {
    return _staticLand.SArray.sequence(Stream, arr);
  },


  /* Given an object (a.k.a map/hash/dictionary) of streams returns a stream of objects.
   * Same as combineArray but for objects.
   */
  combineObject: function combineObject(obj) {
    return _staticLand.SObject.sequence(Stream, obj);
  },


  /* Given a loose basic-stream, that obeys at least following laws:
   *
   *   1. Stream is a function,
   *   2. It accepts one argument, the subscriber function (aka `sink`);
   *
   * returns a valid basic-steam, that obeys all laws of a stream from protocol,
   * and also immune to violations of any of usage laws except:
   *
   *   2. `sink` must be a function
   */
  fromLoose: function fromLoose(looseStream) {
    return function (_sink) {
      var sink = _sink;
      var disposer = looseStream(function (x) {
        // fix usage #1
        if (sink !== null) {
          // fix stream #6
          sink(x); // fix stream #4
        }
        // fix usage #3
      });
      return function () {
        // fix stream #3
        if (sink === null) {
          // fix usage #5
          return;
        }
        sink = null;
        if (typeof disposer === 'function') {
          // fix stream #3
          disposer(); // fix usage #4
        }
        disposer = null;
        // fix stream #5
      };
    };
  },


  /* Given a predicate `A => boolean` returns a function
   * that operates on streams `Stream<A> => Stream<A>`.
   * The result function returns a stream without values that don't satisfy predicate.
   */
  filter: function filter(predicate, stream) {
    return function (sink) {
      return stream(function (payload) {
        if (predicate(payload)) {
          sink(payload);
        }
      });
    };
  },


  /* Same as `chain()`, except the final stream will contain
   * only that values from each spawned streams that was
   * emitted before the next stream was spawned.
   */
  chainLatest: function chainLatest(fn, stream) {
    return function (sink) {
      var spawnedDisposers = function spawnedDisposers() {};
      var mainDisposer = stream(function (payload) {
        spawnedDisposers();
        spawnedDisposers = fn(payload)(sink);
      });
      return function () {
        spawnedDisposers();
        mainDisposer();
      };
    };
  },


  /* Given a reducer function `(B, A) => B`, and a seed value of type B,
   * returns a function that operates on streams `Stream<A> => Stream<B>`.
   * Will apply reducer to each value in `Stream<A>` and push the result to `Stream<B>`
   */
  scan: function scan(reducer, seed, stream) {
    return function (sink) {
      var current = seed;
      sink(current);
      return stream(function (x) {
        current = reducer(current, x);
        sink(current);
      });
    };
  },


  /* Given a number N, returns a function that operates on streams `Stream<A> => Stream<A>`.
   * The result stream will contain only first N items from source stream
   */
  take: function take(n, stream) {
    return function (sink) {
      var count = 0;
      var disposer = null;
      var dispose = function dispose() {
        if (disposer !== null) {
          disposer();
          disposer = null;
        }
      };
      disposer = stream(function (x) {
        count++;
        if (count <= n) {
          sink(x);
        }
        if (count >= n) {
          dispose();
        }
      });
      if (count >= n) {
        dispose();
      }
      return dispose;
    };
  },


  /* Given a predicate `A => boolean` returns a function
   * that operates on streams `Stream<A> => Stream<A>`.
   * The result stream will contain values from source stream
   * before the first value that doesn't satisfy predicate.
   */
  takeWhile: function takeWhile(pred, stream) {
    return function (sink) {
      var completed = false;
      var disposer = null;
      var dispose = function dispose() {
        if (disposer !== null) {
          disposer();
          disposer = null;
        }
      };
      disposer = stream(function (x) {
        if (!completed) {
          if (pred(x)) {
            sink(x);
          } else {
            completed = true;
            dispose();
          }
        }
      });
      if (completed) {
        dispose();
      }
      return dispose;
    };
  },


  /* Given a controller stream, returns a function that operates
   * on streams `Stream<A> => Stream<A>`. The result stream will contain values
   * from source stream until the first value from controller stream.
   */
  takeUntil: function takeUntil(controller, stream) {
    return function (sink) {
      var mainDisposer = null;
      var ctrlDisposer = null;
      var dispose = function dispose() {
        if (mainDisposer !== null) {
          mainDisposer();
          mainDisposer = null;
        }
        if (ctrlDisposer !== null) {
          ctrlDisposer();
          ctrlDisposer = null;
        }
      };

      mainDisposer = stream(sink);
      ctrlDisposer = controller(dispose);

      // in case controller cb was called sync before we obtained ctrlDisposer
      if (mainDisposer === null) {
        dispose();
      }

      return dispose;
    };
  },


  /* Given a number N, returns a function that operates on streams `Stream<A> => Stream<A>`.
   * The result stream will contain only items from source starting from (N+1)th one
   */
  skip: function skip(n, stream) {
    return function (sink) {
      var count = 0;
      return stream(function (x) {
        count++;
        if (count > n) {
          sink(x);
        }
      });
    };
  },


  /* Given a predicate `A => boolean` returns a function
   * that operates on streams `Stream<A> => Stream<A>`.
   * The result stream will contain values from source stream
   * starting from the first that doesn't satisfy predicate.
   */
  skipWhile: function skipWhile(pred, stream) {
    return function (sink) {
      var started = false;
      return stream(function (x) {
        if (!started) {
          started = !pred(x);
        }
        if (started) {
          sink(x);
        }
      });
    };
  },


  /* Given a comparator function `(A, A) => boolean` returns a function that
   * operates on streams `Stream<A> => Stream<A>`. For each element X from
   * the source stream (except the first one) calls comparator with (Y, X)
   * as arguments, where Y is the last element in the result stream.
   * If comparator returns false, X goes to the result stream.
   */
  skipDuplicates: function skipDuplicates(comp, stream) {
    return function (sink) {
      var latest = { type: 'nothing' };
      return stream(function (x) {
        if (latest.type === 'nothing' || !comp(latest.value, x)) {
          latest = { type: 'just', value: x };
          sink(x);
        }
      });
    };
  },


  /* Given a stream returns a new stream of same type. The new stream will
   * have at most one subscription at any given time to the original stream.
   * It allows you to connect several subscribers to a stream using only one subscription.
   */
  multicast: function multicast(stream) {
    var sinks = [];
    var push = function push(x) {
      sinks.forEach(function (sink) {
        if (sinks.indexOf(sink) !== -1) {
          sink(x);
        }
      });
    };
    var unsub = null;
    return function (sink) {
      var disposed = false;
      sinks = [].concat(_toConsumableArray(sinks), [sink]);
      if (sinks.length === 1) {
        unsub = stream(push);
      }
      return function () {
        if (disposed) {
          return;
        }
        disposed = true;
        var index = sinks.indexOf(sink);
        sinks = [].concat(_toConsumableArray(sinks.slice(0, index)), _toConsumableArray(sinks.slice(index + 1, sinks.length)));
        if (sinks.length === 0 && unsub !== null) {
          unsub();
          unsub = null;
        }
      };
    };
  },


  /* Given a value of type `A` returns a function that
   * operates on streams `Stream<A> => Stream<A>`.
   * The result stream is a copy of source stream
   * with the given value added at the beginning.
   */
  startWith: function startWith(x, stream) {
    return function (sink) {
      sink(x);
      return stream(sink);
    };
  },


  /* Given a transducer that conforms [Protocol][1],
   * returns a function that operates on streams `Stream<A> => Stream<B>`.
   *
   * [1]: https://github.com/cognitect-labs/transducers-js#the-transducer-protocol
   */
  transduce: function transduce(transducer, stream) {
    return function (sink) {
      var thisDisposed = false;
      var sourceDisposer = null;

      var transformer = transducer({
        '@@transducer/result': function transducerResult() {},
        '@@transducer/step': function transducerStep(result, input) {
          if (!thisDisposed) {
            sink(input);
          }
          return result;
        }
      });

      var disposeSource = function disposeSource() {
        if (sourceDisposer !== null) {
          sourceDisposer();
          sourceDisposer = null;
        }
      };

      sourceDisposer = stream(function (x) {
        if (transformer === null) {
          return;
        }

        // it returns either null or tReduced<null>
        if (null !== transformer['@@transducer/step'](null, x)) {
          transformer['@@transducer/result'](null);
          transformer = null;
          disposeSource();
        }
      });

      if (transformer === null) {
        disposeSource();
      }

      return function () {
        thisDisposed = true;
        disposeSource();
      };
    };
  },


  composeWithInnerType: _compose2.default

});

exports.default = Stream;